"""The main parsing routine."""

import inspect
from typing import Optional

from pymend.docstring_parser import epydoc, google, numpydoc, rest
from pymend.docstring_parser.attrdoc import add_attribute_docstrings
from pymend.docstring_parser.common import (
    Docstring,
    DocstringStyle,
    ParseError,
    RenderingStyle,
)

_STYLE_MAP = {
    DocstringStyle.REST: rest,
    DocstringStyle.GOOGLE: google,
    DocstringStyle.NUMPYDOC: numpydoc,
    DocstringStyle.EPYDOC: epydoc,
}


def parse(
    text: Optional[str], style: DocstringStyle = DocstringStyle.AUTO
) -> Docstring:
    """Parse the docstring into its components.

    Parameters
    ----------
    text : Optional[str]
        docstring text to parse
    style : DocstringStyle
        docstring style (Default value = DocstringStyle.AUTO)

    Returns
    -------
    Docstring
        parsed docstring representation

    Raises
    ------
    ParserError
        If none of the available module an parse the docstring
    """
    if style != DocstringStyle.AUTO:
        return _STYLE_MAP[style].parse(text)

    exc: Optional[Exception] = None
    rets: list[Docstring] = []
    for module in _STYLE_MAP.values():
        try:
            ret = module.parse(text)
        except ParseError as ex:
            exc = ex
        else:
            rets.append(ret)

    if not rets and exc:
        raise exc

    return sorted(rets, key=lambda d: (len(d.examples), len(d.meta)), reverse=True)[0]


def parse_from_object(
    obj: object,
    style: DocstringStyle = DocstringStyle.AUTO,
) -> Docstring:
    """Parse the object's docstring(s) into its components.

    The object can be anything that has a ``__doc__`` attribute. In contrast to
    the ``parse`` function, ``parse_from_object`` is able to parse attribute
    docstrings which are defined in the source code instead of ``__doc__``.

    Currently only attribute docstrings defined at class and module levels are
    supported. Attribute docstrings defined in ``__init__`` methods are not
    supported.

    When given a class, only the attribute docstrings of that class are parsed,
    not its inherited classes. This is a design decision. Separate calls to
    this function should be performed to get attribute docstrings of parent
    classes.

    Parameters
    ----------
    obj : object
        object from which to parse the docstring(s)
    style : DocstringStyle
        docstring style (Default value = DocstringStyle.AUTO)

    Returns
    -------
    Docstring
        parsed docstring representation
    """
    docstring = parse(obj.__doc__, style=style)

    if inspect.isclass(obj) or inspect.ismodule(obj):
        add_attribute_docstrings(obj, docstring)

    return docstring


def compose(
    docstring: Docstring,
    style: DocstringStyle = DocstringStyle.AUTO,
    rendering_style: RenderingStyle = RenderingStyle.COMPACT,
    indent: str = "    ",
) -> str:
    """Render a parsed docstring into docstring text.

    Parameters
    ----------
    docstring : Docstring
        parsed docstring representation
    style : DocstringStyle
        docstring style to render (Default value = DocstringStyle.AUTO)
    indent : str
        the characters used as indentation in the docstring string
        (Default value = '    ')
    rendering_style : RenderingStyle
        The rendering style to use. (Default value = RenderingStyle.COMPACT)

    Returns
    -------
    str
        docstring text

    Raises
    ------
    ValueError
        If no output style could be determined.
    """
    if style == DocstringStyle.AUTO:
        if docstring.style is None:
            msg = (
                "Detected docstring.style of `None` and requested style of `AUTO`.\n"
                "Either the docstring to compose has to have its style set"
                " (for example by calling `parse`) or an "
                "output style has to be provided."
            )
            raise ValueError(msg)
        style = docstring.style
    module = _STYLE_MAP[style]
    return module.compose(docstring, rendering_style=rendering_style, indent=indent)



"""Testing issues raised on github."""

import os

import pytest

import pymend.pymend as pym
from pymend.types import FixerSettings

current_dir = os.path.dirname(__file__)


def absdir(file: str) -> str:
    """Get absolute path for file.

    Parameters
    ----------
    file : str
        File path

    Returns
    -------
    str
        Absolute path to file
    """
    return os.path.join(current_dir, file)


class TestIssues:
    """Class for testing raised issues."""

    def test_issue_30(self) -> None:
        """Test issue 30.

        https://github.com/dadadel/pymend/issues/30
        """
        # if file starting with a function/class definition, patching the file
        # will remove the first line!
        comment = pym.PyComment(
            absdir("refs/issue30.py"), fixer_settings=FixerSettings()
        )
        try:
            comment._docstring_diff()
        except Exception as e:  # noqa: BLE001
            pytest.fail(f'Raised exception: "{e}"')

    def test_issue_49(self) -> None:
        """Test issue 49.

        https://github.com/dadadel/pymend/issues/49
        """
        # Title: If already numpydoc format, will remove the Raises section
        # If the last section in a numpydoc docstring is a `Raises` section,
        # it will be removed if the output format is also set to numpydoc
        comment = pym.PyComment(
            absdir("refs/issue49.py"), fixer_settings=FixerSettings()
        )
        result = "".join(comment._docstring_diff())
        assert result == ""



"""_summary_."""
def my_func(param0, param01: int, param1: str = "Some value", param2: List[str] = {}):
    """_summary_.

    Args:
        param0 (_type_): _description_
        param01 (int): _description_
        param1 (str, optional): _description_. Defaults to "Some value".
        param2 (List[str], optional): _description_. Defaults to {}.
    """
    pass


def my_single_return_func1() -> str:
    """_summary_.

    Returns
    -------
    int
        Wrong
    """
    pass


def my_single_return_func2():
    """_summary_.

    Returns
    -------
    int
        Wrong
    """
    pass


def my_single_return_func3():
    pass

def my_single_return_func4():
    """Existing docstring."""
    pass

def my_single_return_func5() -> None:
    """Existing docstring."""
    pass

def my_single_return_func6() -> None:
    """Existing docstring."""
    return None

def my_single_return_func7():
    """Existing docstring."""
    return None


def my_func1(param0, param01: int):
    """_summary_.

    Args:
        param0 (_type_): _description_
        param01 (int): _description_
    """
    pass


def my_multi_return_func() -> Tuple[int, str, bool]:
    """_summary_.

    Returns
    -------
    x :
        Some integer
    y : str
        Some string
    z : bool
        Some bool
    """
    pass



"""_summary_."""
def func1(param1):
    """Function to test issue #49.

    Parameters
    ----------
    param1 : object
        Description of param1

    Returns
    -------
    None
        Description for return value

    Raises
    ------
    AssertionError
        Raises an error if run
    """
    assert False



"""Numpydoc-style docstring parsing.

See
---
https://numpydoc.readthedocs.io/en/latest/format.html
"""

import inspect
import itertools
import re
from collections.abc import Iterable, Iterator
from textwrap import dedent
from typing import Optional, TypeVar

from typing_extensions import override

from .common import (
    Docstring,
    DocstringDeprecated,
    DocstringExample,
    DocstringMeta,
    DocstringParam,
    DocstringRaises,
    DocstringReturns,
    DocstringStyle,
    DocstringYields,
    MainSections,
    ParseError,
    RenderingStyle,
    clean_str,
    split_description,
)

_T = TypeVar("_T")


def _pairwise(
    iterable: Iterable[_T], end: Optional[_T] = None
) -> Iterator[tuple[_T, Optional[_T]]]:
    """Iterate over successive pairs with overhang for last element.

    Parameters
    ----------
    iterable : Iterable[_T]
        Iterable to iterate over.
    end : Optional[_T]
        Value for the overhang (Default value = None)

    Returns
    -------
    Iterator[tuple[_T, Optional[_T]]]
        Iterator yielding the successive pairs.
    """
    left, right = itertools.tee(iterable)
    next(right, None)
    return zip(left, itertools.chain(right, [end]))


KV_REGEX = re.compile(r"^[^\s].*$", flags=re.M)
PARAM_KEY_REGEX = re.compile(r"^(?P<name>.*?)(?:\s+:\s*(?P<type>.*?))?$")
PARAM_OPTIONAL_REGEX = re.compile(r"(?P<type>.*?)(?:, optional|\(optional\))$")

# numpydoc format has no formal grammar for this,
# but we can make some educated guesses...
PARAM_DEFAULT_REGEX = re.compile(
    r"(?<!\S)[Dd]efault(?: is | = |: |s to |)\s*(?P<value>[\w\-\.]*\w)"
)

RETURN_KEY_REGEX = re.compile(r"^(?:(?P<name>.*?)\s*:\s*)?(?P<type>.*?)$")


class Section:
    """Numpydoc section parser."""

    def __init__(self, title: str, key: str) -> None:
        """Initialize a section.

        Parameters
        ----------
        title : str
            section title. For most sections, this is a heading like
            "Parameters" which appears on its own line, underlined by
            en-dashes ('-') on the following line.
        key : str
            meta key string. In the parsed ``DocstringMeta`` instance this
            will be the first element of the ``args`` attribute list.
        """
        self.title = title
        self.key = key

    @property
    def title_pattern(self) -> str:
        """Regular expression pattern matching this section's header.

        This pattern will match this instance's ``title`` attribute in
        an anonymous group.

        Returns
        -------
        str
            Regex pattern as a string.
        """
        dashes = "-" * len(self.title)
        return rf"^({self.title})\s*?\n{dashes}\s*$"

    def parse(self, text: str) -> Iterable[DocstringMeta]:
        """Parse ``DocstringMeta`` objects from the body of this section.

        Parameters
        ----------
        text : str
            section body text. Should be cleaned with
            ``inspect.cleandoc`` before parsing.

        Yields
        ------
        DocstringMeta
            object from this section body.
        """
        yield DocstringMeta([self.key], description=clean_str(text))


class _KVSection(Section):
    """Base parser for numpydoc sections with key-value syntax.

    E.g. sections that look like this:
        key
            value
        key2 : type
            values can also span...
            ... multiple lines
    """

    def _parse_item(self, key: str, value: str) -> DocstringMeta:
        """Abstract method for parsing a single item of a section.

        Parameters
        ----------
        key : str
            Key of the item to parse
        value : str
            Value of the item to parse

        Raises
        ------
        NotImplementedError
            To be implemented by child classes.
        """
        raise NotImplementedError

    @override
    def parse(self, text: str) -> Iterable[DocstringMeta]:
        """Parse all items in the docstring text.

        Parameters
        ----------
        text : str
            Docstring text to parse.

        Yields
        ------
        DocstringMeta
            Items parsed from the docstring.
        """
        for match, next_match in _pairwise(KV_REGEX.finditer(text)):
            start = match.end()
            end = next_match.start() if next_match is not None else None
            value = text[start:end]
            yield self._parse_item(key=match.group(), value=inspect.cleandoc(value))


class _SphinxSection(Section):
    """Base parser for numpydoc sections with sphinx-style syntax.

    E.g. sections that look like this:
        .. title:: something
            possibly over multiple lines
    """

    @property
    @override
    def title_pattern(self) -> str:
        """Title pattern used by sphinx sections.

        Returns
        -------
        str
            Regex pattern as a string.
        """
        return rf"^\.\.\s*({self.title})\s*::"


class ParamSection(_KVSection):
    """Parser for numpydoc parameter sections.

    E.g. any section that looks like this:
        arg_name
            arg_description
        arg_2 : type, optional
            descriptions can also span...
            ... multiple lines
    """

    @override
    def _parse_item(self, key: str, value: str) -> DocstringParam:
        """Parse item from a parameter section.

        Parameters
        ----------
        key : str
            Key of the item. Contains parameter name and optionally type information.
        value : str
            Description for the item. Also possibly contains default value.

        Returns
        -------
        DocstringParam
            Parsed representation of the parameter item.

        Raises
        ------
        ParseError
            If no key could be parsed.
        ParseError
            If mandatory parts of the section were parsed incorrectly.
        """
        match = PARAM_KEY_REGEX.match(key)
        arg_name = type_name = is_optional = None
        if match is None:
            msg = f"Could not parse param key on line `{key}`"
            raise ParseError(msg)
        arg_name = match.group("name")
        type_name = match.group("type")
        if not isinstance(arg_name, str):
            msg = (
                f"Did not get a string when capturing mandatory section"
                f" 'arg_name' for key line `{key}`. Got `{arg_name}` instead."
            )
            raise ParseError(msg)
        if isinstance(type_name, str):
            optional_match = PARAM_OPTIONAL_REGEX.match(type_name)
            if optional_match is not None:
                type_name = optional_match.group("type")
                is_optional = True
            else:
                is_optional = False
        else:
            type_name = None

        default = None
        if value != "":
            default_match = PARAM_DEFAULT_REGEX.search(value)
            if default_match is not None:
                default = default_match.group("value")

        return DocstringParam(
            args=[self.key, arg_name],
            description=clean_str(value),
            arg_name=arg_name,
            type_name=type_name,
            is_optional=is_optional,
            default=default,
        )


class RaisesSection(_KVSection):
    """Parser for numpydoc raises sections.

    E.g. any section that looks like this:
        ValueError
            A description of what might raise ValueError
    """

    @override
    def _parse_item(self, key: str, value: str) -> DocstringRaises:
        """Parse an item in the raises section.

        Parameters
        ----------
        key : str
            Key of the item to be parsed. Usually name of the exception raised.
        value : str
            Description of the item.

        Returns
        -------
        DocstringRaises
            Parsed representation of the raises item.
        """
        return DocstringRaises(
            args=[self.key, key],
            description=clean_str(value),
            type_name=key if key != "" else None,
        )


class ReturnsSection(_KVSection):
    """Parser for numpydoc returns sections.

    E.g. any section that looks like this:
        return_name : type
            A description of this returned value
        another_type
            Return names are optional, types are required
    """

    is_generator = False

    @override
    def _parse_item(self, key: str, value: str) -> DocstringReturns:
        """Parse an item from the return section.

        Parameters
        ----------
        key : str
            Key of the item (usually type, possibly name + type)
        value : str
            Description of the return value.

        Returns
        -------
        DocstringReturns
            Parsed representation of the return item.
        """
        match = RETURN_KEY_REGEX.match(key)
        if match is not None:
            return_name = match.group("name")
            type_name = match.group("type")
        else:
            return_name = None
            type_name = None

        return DocstringReturns(
            args=[self.key],
            description=clean_str(value),
            type_name=type_name,
            is_generator=self.is_generator,
            return_name=return_name,
        )


class YieldsSection(_KVSection):
    """Parser for numpydoc generator "yields" sections."""

    is_generator = True

    @override
    def _parse_item(self, key: str, value: str) -> DocstringYields:
        """Parse an item from the yield section.

        Parameters
        ----------
        key : str
            Key of the item (usually type, possibly name + type)
        value : str
            Description of the yielded value.

        Returns
        -------
        DocstringYields
            Parsed representation of the yield item.
        """
        match = RETURN_KEY_REGEX.match(key)
        if match is not None:
            yield_name = match.group("name")
            type_name = match.group("type")
        else:
            yield_name = None
            type_name = None

        return DocstringYields(
            args=[self.key],
            description=clean_str(value),
            type_name=type_name,
            is_generator=self.is_generator,
            yield_name=yield_name,
        )


class DeprecationSection(_SphinxSection):
    """Parser for numpydoc "deprecation warning" sections.

    E.g. any section that looks like this:
        .. deprecated:: 1.6.0
            This description has
            multiple lines!
    """

    @override
    def parse(self, text: str) -> Iterable[DocstringDeprecated]:
        """Parse ``DocstringDeprecated`` objects from the body of this section.

        Parameters
        ----------
        text : str
            Text of the deprecation section.

        Yields
        ------
        DocstringDeprecated
            Parsed representation of the deprecation item.

        Raises
        ------
        ParseError
            If the parsed version number was unexpectedly `None`.
            Usually a lack of version number would be represented by an empty string.
        """
        version, desc, *_ = [*text.split(sep="\n", maxsplit=1), None, None]
        if version is None:
            msg = (
                f"Got `None` while parsing version number "
                f"in deprecated section `{text}`."
            )
            raise ParseError(msg)
        if desc is not None:
            desc = clean_str(inspect.cleandoc(desc))

        yield DocstringDeprecated(
            args=[self.key], description=desc, version=clean_str(version)
        )


class ExamplesSection(Section):
    """Parser for numpydoc examples sections.

    E.g. any section that looks like this:

        Optional description for the following example. Always preceded
        and followed by an empty line. Except for the first description.

        >>> import numpy.matlib
        >>> np.matlib.empty((2, 2))    # filled with random data
        matrix([[  6.76425276e-320,   9.79033856e-307], # random
                [  7.39337286e-309,   3.22135945e-309]])

        Description for the second example.

        >>> d = np.zeros((5,2))
        >>> for i in range(5):
        ...   for j in range(2):
        ...     for k in range(3):
        ...       for n in range(4):
        ...         d[i,j] += a[k,n,i] * b[n,k,j]
        >>> c == d
        array([[ True,  True],
            [ True,  True],
            [ True,  True],
            [ True,  True],
            [ True,  True]])
    """

    @override
    def parse(self, text: str) -> Iterable[DocstringExample]:
        """Parse ``DocstringExample`` objects from the body of this section.

        Parameters
        ----------
        text : str
            section body text. Should be cleaned with
            ``inspect.cleandoc`` before parsing.

        Yields
        ------
        DocstringExample
            Docstring example sections
        """
        # Reverse so that we can efficiently pop from the back
        # instead of doing constant pops from the front.
        # Could also use a deque
        # ---
        # We add a newline to the end to not have to special case the first
        # description.
        lines = [*list(reversed(dedent(text).strip().splitlines())), "\n"]
        while lines:
            snippet_lines: list[str] = []
            description_lines: list[str] = []
            # Empty lines before the description
            while lines and lines[-1].strip() == "":
                lines.pop()
            # Description. Should not start with ">>>". if that were the case
            # Then there was no description.
            while lines and lines[-1].strip() != "" and not lines[-1].startswith(">>>"):
                description_lines.append(lines.pop())
            # Empty lines after description
            while lines and lines[-1].strip() == "":
                lines.pop()
            # Here the actual example starts.
            # We take any line.
            # The code part starts with ">>>" or "..."
            # but the result part can be anything.
            # Just keeping until an empty line which should indicate the next example.
            while lines and lines[-1].strip() != "":
                snippet_lines.append(lines.pop())
            yield DocstringExample(
                [self.key],
                snippet="\n".join(snippet_lines) if snippet_lines else None,
                description="\n".join(description_lines),
            )


DEFAULT_SECTIONS = [
    ParamSection("Parameters", "param"),
    ParamSection("Params", "param"),
    ParamSection("Arguments", "param"),
    ParamSection("Args", "param"),
    ParamSection("Other Parameters", "other_param"),
    ParamSection("Other Params", "other_param"),
    ParamSection("Other Arguments", "other_param"),
    ParamSection("Other Args", "other_param"),
    ParamSection("Receives", "receives"),
    ParamSection("Receive", "receives"),
    RaisesSection("Raises", "raises"),
    RaisesSection("Raise", "raises"),
    RaisesSection("Warns", "warns"),
    RaisesSection("Warn", "warns"),
    ParamSection("Attributes", "attribute"),
    ParamSection("Attribute", "attribute"),
    ParamSection("Methods", "method"),
    ParamSection("Method", "method"),
    ReturnsSection("Returns", "returns"),
    ReturnsSection("Return", "returns"),
    YieldsSection("Yields", "yields"),
    YieldsSection("Yield", "yields"),
    ExamplesSection("Examples", "examples"),
    ExamplesSection("Example", "examples"),
    Section("Warnings", "warnings"),
    Section("Warning", "warnings"),
    Section("See Also", "see_also"),
    Section("Related", "see_also"),
    Section("Notes", "notes"),
    Section("Note", "notes"),
    Section("References", "references"),
    Section("Reference", "references"),
    DeprecationSection("deprecated", "deprecation"),
]


class NumpydocParser:
    """Parser for numpydoc-style docstrings."""

    def __init__(self, sections: Optional[Iterable[Section]] = None) -> None:
        """Set up sections.

        Parameters
        ----------
        sections : Optional[Iterable[Section]]
            Recognized sections or None to defaults.
        """
        self.sections = {s.title: s for s in (sections or DEFAULT_SECTIONS)}
        self._setup()

    def _setup(self) -> None:
        """Set up parser title regex."""
        self.titles_re = re.compile(
            r"|".join(s.title_pattern for s in self.sections.values()),
            flags=re.M,
        )

    def add_section(self, section: Section) -> None:
        """Add or replace a section.

        Parameters
        ----------
        section : Section
            The new section.
        """
        self.sections[section.title] = section
        self._setup()

    def parse(self, text: Optional[str]) -> Docstring:
        """Parse the numpy-style docstring into its components.

        Parameters
        ----------
        text : Optional[str]
            docstring text

        Returns
        -------
        Docstring
            parsed docstring
        """
        ret = Docstring(style=DocstringStyle.NUMPYDOC)
        if not text:
            return ret

        # Clean according to PEP-0257
        text = inspect.cleandoc(text)

        if match := self.titles_re.search(text):
            desc_chunk = text[: match.start()]
            meta_chunk = text[match.start() :]
        else:
            desc_chunk = text
            meta_chunk = ""

        # Break description into short and long parts
        split_description(ret, desc_chunk)

        for match, nextmatch in _pairwise(self.titles_re.finditer(meta_chunk)):
            title = next(g for g in match.groups() if g is not None)
            factory = self.sections[title]

            # section chunk starts after the header,
            # ends at the start of the next header
            start = match.end()
            end = nextmatch.start() if nextmatch is not None else None
            ret.meta.extend(factory.parse(meta_chunk[start:end]))

        return ret


def parse(text: Optional[str]) -> Docstring:
    """Parse the numpy-style docstring into its components.

    Parameters
    ----------
    text : Optional[str]
        docstring text

    Returns
    -------
    Docstring
        parsed docstring
    """
    return NumpydocParser().parse(text)


def process_examples(examples: list[DocstringExample], parts: list[str]) -> None:
    """Add string representation of examples section to parts.

    Parameters
    ----------
    examples : list[DocstringExample]
        DocstringExamples to add to parts.
    parts : list[str]
        List of strings representing the final output of compose().
    indent : str
        the characters used as indentation in the docstring string
        (Default value = '    ')
    """
    if examples:
        parts.append("")
        parts.append("Examples")
        parts.append("-" * len(parts[-1]))
        for i, example in enumerate(examples):
            # Leave out newline for first example
            if i != 0:
                parts.append("\n")
            if example.description:
                parts.append("\n".join(example.description.splitlines()))
                # Only add a new line if we have an actual example snippet here.
                # If not the next description will handle it.
                if example.snippet:
                    parts.append("\n")
            if example.snippet:
                parts.append("\n".join(example.snippet.splitlines()))


def compose(  # noqa: PLR0915
    # pylint: disable=W0613,R0915
    docstring: Docstring,
    rendering_style: RenderingStyle = RenderingStyle.COMPACT,  # noqa: ARG001
    indent: str = "    ",
) -> str:
    """Render a parsed docstring into docstring text.

    Parameters
    ----------
    docstring : Docstring
        parsed docstring representation
    rendering_style : RenderingStyle
        the style to render docstrings (Default value = RenderingStyle.COMPACT)
    indent : str
        the characters used as indentation in the docstring string
        (Default value = '    ')

    Returns
    -------
    str
        docstring text
    """

    def process_one(one: MainSections) -> None:
        """Build the output text for one entry in a section.

        Parameters
        ----------
        one : MainSections
            Docstring for which to build the raw text.
        """
        if isinstance(one, DocstringParam):
            head = one.arg_name
        elif isinstance(one, DocstringReturns):
            head = one.return_name
        elif isinstance(one, DocstringYields):
            head = one.yield_name
        else:
            head = None

        if one.type_name and head:
            head += f" : {one.type_name}"
        elif one.type_name:
            head = one.type_name
        elif not head:
            head = ""

        if isinstance(one, DocstringParam) and one.is_optional:
            head += ", optional"

        if one.description:
            body = f"\n{indent}".join([head, *one.description.splitlines()])
            parts.append(body)
        else:
            parts.append(head)

    def process_sect(name: str, args: list[MainSections]) -> None:
        """Build the output for a docstring section.

        Parameters
        ----------
        name : str
            Section for which to build the output.
        args : list[MainSections]
            List of individual elements of that section.
        """
        if args:
            parts.append("")
            parts.append(name)
            parts.append("-" * len(name))
            for arg in args:
                process_one(arg)

    parts: list[str] = []
    if docstring.short_description:
        parts.append(docstring.short_description)
    if docstring.blank_after_short_description:
        parts.append("")

    if docstring.deprecation:
        first = ".. deprecated::"
        if docstring.deprecation.version:
            first += f" {docstring.deprecation.version}"
        if docstring.deprecation.description:
            rest = docstring.deprecation.description.splitlines()
        else:
            rest = []
        sep = f"\n{indent}"
        parts.append(sep.join([first, *rest]))

    if docstring.long_description:
        parts.append(docstring.long_description)
    if docstring.blank_after_long_description:
        parts.append("")

    process_sect(
        "Parameters",
        [item for item in docstring.params or [] if item.args[0] == "param"],
    )

    process_sect(
        "Attributes",
        [item for item in docstring.params or [] if item.args[0] == "attribute"],
    )

    process_sect(
        "Methods",
        [item for item in docstring.params or [] if item.args[0] == "method"],
    )

    process_sect(
        "Returns",
        list(docstring.many_returns or []),
    )

    process_sect(
        "Yields",
        list(docstring.many_yields or []),
    )

    if docstring.returns and not docstring.many_returns:
        ret = docstring.returns
        parts.append("Yields" if ret else "Returns")
        parts.append("-" * len(parts[-1]))
        process_one(ret)

    process_sect(
        "Receives",
        [item for item in docstring.params or [] if item.args[0] == "receives"],
    )

    process_sect(
        "Other Parameters",
        [item for item in docstring.params or [] if item.args[0] == "other_param"],
    )

    process_sect(
        "Raises",
        [item for item in docstring.raises or [] if item.args[0] == "raises"],
    )

    process_sect(
        "Warns",
        [item for item in docstring.raises or [] if item.args[0] == "warns"],
    )

    process_examples(docstring.examples, parts)

    for meta in docstring.meta:
        if isinstance(
            meta,
            (
                DocstringDeprecated,
                DocstringParam,
                DocstringReturns,
                DocstringRaises,
                DocstringYields,
                DocstringExample,
            ),
        ):
            continue  # Already handled

        parts.append("")
        parts.append(meta.args[0].replace("_", "").title())
        parts.append("-" * len(meta.args[0]))

        if meta.description:
            parts.append(meta.description)

    return "\n".join(parts)



"""Integration tests of pymend app."""

import os
import re
import subprocess
import tempfile
import textwrap
from re import Pattern
from typing import Optional, Union

import pytest

import pymend.pymend


def remove_diff_header(diff: str) -> str:
    """Remove header differences from diff.

    Parameters
    ----------
    diff : str
        Diff file to clean.

    Returns
    -------
    str
        Cleaned diff.
    """
    return re.sub(r"(@@.+@@)|(\-\-\-.*)|(\+\+\+.*)", "", diff)


class TestApp:
    """Test pymend as an app in a shell.

    It's an integration test.
    """

    def setup_class(self) -> None:
        """Set up class by defining docstrings."""
        # You have to run this as a module when testing so the relative imports work.
        self.CMD_PREFIX = "pymend {}"

        self.RE_TYPE = type(re.compile("get the type to test if an argument is an re"))

        # cwd to use when running subprocess.
        # It has to be at the repo directory so python -m can be used
        self.CWD = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))

        self.INPUT = textwrap.dedent(
            '''

            def func():
                """First line

                :returns: smthg

                :rtype: ret type

                """
                pass
        '''
        )

        # Expected output in overwrite mode.
        self.EXPECTED_OUTPUT = textwrap.dedent(
            '''\
            """_summary_."""

            def func():
                """First line.

                Returns
                -------
                ret type
                    smthg
                """
                pass
        '''
        )

        self.PATCH_PREFIX = f"# Patch generated by Pymend v{pymend.pymend.__version__}"

        # a/- and b/- is replaced by a filename when not testing stdin/stdout
        self.EXPECTED_PATCH = textwrap.dedent(
            f'''\
            {self.PATCH_PREFIX}

            --- a/-
            +++ b/-
            @@ -1,11 +1,12 @@
            +"""_summary_."""

             def func():
            -    """First line
            +    """First line.

            -    :returns: smthg
            -
            -    :rtype: ret type
            -
            +    Returns
            +    -------
            +    ret type
            +        smthg
                 """
                 pass

        '''
        )

    @classmethod
    def normalise_empty_lines(cls, lines: str) -> str:
        r"""Replace any lines that are only whitespace with a single '\n'.

        textwrap.dedent removes all whitespace characters
        on lines only containing whitespaces
        see: https://bugs.python.org/issue30754

        And some people set their editors to strip trailing white space.

        But sometimes there is a space on an empty line in the
        output which will fail the comparison.

        So strip the spaces on empty lines

        Parameters
        ----------
        lines : str
            string of lines to normalise

        Returns
        -------
        str
            normalised lines
        """
        return re.sub(r"^\s+$", "", lines, flags=re.MULTILINE)

    def run_command(
        self, cmd_to_run: str, write_to_stdin: Optional[str] = None
    ) -> tuple[str, str, int]:
        r"""Run a command in shell mode returning stdout, stderr and the returncode.

        Parameters
        ----------
        cmd_to_run : str
            shell command to run
        write_to_stdin : str | None
            string to put on stdin if not None (Default value = None)

        Returns
        -------
        tuple[str, str, int]
            stdout, stderr, returncode
        """
        process = subprocess.Popen(
            cmd_to_run,
            shell=True,  # noqa: S602
            cwd=self.CWD,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if write_to_stdin:
            # Python3 compatibility - input has to be bytes
            write_to_stdin = write_to_stdin.encode()
        stdout, stderr = process.communicate(write_to_stdin)

        stdout = stdout.decode()
        stderr = stderr.decode()
        return stdout, stderr, process.returncode

    def run_pymend_app_and_assert_is_expected(
        self,
        cmd_args: str,
        write_to_stdin: Optional[str] = None,
        expected_stdout: Union[str, Pattern[str]] = "",
        expected_stderr: Union[str, Pattern[str]] = "",
        expected_returncode: int = 0,
    ) -> None:
        """Run pymend and assert it's output matches the arguments.

        With the cmd_args and output_format specified in a shell.

        If expected_stdout and expected_stderr is the result of a re.compile()
        the output will be checked with re.search().

        Parameters
        ----------
        cmd_args : str
            Extra arguments to pass to pymend - excluding the output_format
        write_to_stdin : Optional[str]
            The input to put on stdin,
            use None if there's nothing (Default value = None)
        expected_stdout : str | Pattern[str]
            Expected string to see on stdout (Default value = "")
        expected_stderr : str | Pattern[str]
            Expected string to see on stderr (Default value = "")
        expected_returncode : int
            Expected returncode after running pymend (Default value = 0)

        Raises
        ------
        AssertionError
            If the expected result is not found
        """

        def assert_output(
            cmd_to_run: str,
            what: str,
            got: Union[str, int],
            expected: Union[str, Pattern[str], int],
        ) -> None:
            """See run_pymend_app_and_assert_is_expected.

            Parameters
            ----------
            cmd_to_run : str
                full command that was run - used to build an error message
            what : str
                The attribute being checked - used for the error message
            got : str
                The result from the test
            expected : Union[str, Pattern[str]]
                The expected result from the test

            Raises
            ------
            AssertionError
                If the expected result is not found
            """
            if isinstance(expected, self.RE_TYPE):
                msg = (
                    f"Test failed for cmd {cmd_to_run}\n{what} was "
                    f"expected to match the regex:\n{expected}\n"
                    f"But this was the output:\n{got!r}\n"
                )
                assert expected.search(got) is not None, msg
            else:
                if isinstance(expected, str):
                    # Turn lines that only have whitespace into
                    # single newline lines to workaround textwrap.dedent
                    # behaviour
                    got = remove_diff_header(
                        self.normalise_empty_lines(got).replace("\r\n", "\n")
                    )
                    expected = remove_diff_header(self.normalise_empty_lines(expected))
                # repr is used instead of str to make it easier
                # to see newlines and spaces if there's a difference
                msg = (
                    f"Test failed for cmd {cmd_to_run}\n{what} was expected to be:"
                    f"\n{expected!r}\nBut this was the output:\n{got!r}\n"
                )
                assert got == expected, msg

        cmd_to_run = self.CMD_PREFIX.format(cmd_args)

        stdout, stderr, returncode = self.run_command(cmd_to_run, write_to_stdin)

        assert_output(cmd_to_run, "stderr", stderr, expected_stderr)
        assert_output(cmd_to_run, "returncode", returncode, expected_returncode)
        assert_output(cmd_to_run, "stdout", stdout, expected_stdout)

    def test_no_args_ge_py33(self) -> None:
        """Ensure the app outputs an error if there are no arguments."""
        self.run_pymend_app_and_assert_is_expected(
            cmd_args="",
            write_to_stdin=None,
            expected_stderr=re.compile(
                r"Usage: pymend \[OPTIONS\] SRC \.\.\..*"
                r"Error: Missing argument 'SRC \.\.\.'\.",
                re.DOTALL,
            ),
            expected_returncode=1,
        )

    def run_pymend_app_with_a_file_and_assert_is_expected(
        self,
        file_contents: str,
        cmd_args: str = "",
        *,
        overwrite_mode: bool = False,
        expected_file_contents: Union[str, Pattern[str]] = "",
        expected_stderr: Union[str, Pattern[str]] = "",
        expected_stdout: Union[str, Pattern[str]] = "",
        expected_returncode: int = 0,
    ) -> None:
        """
        Run the pymend app with a file - not stdin.

        A temporary file is created,
        file_contents is written into it then the test is run.
        The .patch and temporary files are removed at the end of the test.

        Parameters
        ----------
        file_contents : str
            write this into the temporary file
        cmd_args : str
            Arguments to pymend - do not put the '-w' argument here -
            it is triggered by overwrite_mode (Default value = "")
        overwrite_mode : bool
            set to True if in overwrite mode (Default value = False)
        expected_file_contents : str | Pattern[str]
            expected result - for a patch file ensure the filename is '-'.
            The '-' (Default value = "")
        expected_stderr : str | Pattern[str]
            expected output on stderr.
            You can match on a regex if you pass it the result of (Default value = "")
        expected_stdout : str | Pattern[str]
            Expected string to see on stdout (Default value = "")
        expected_returncode : int
            Expected return code from pymend. Default is 0.
        output_format : Optional[str]
            If not using auto mode set the output format to this. (Default value = None)
        """
        patch_filename = input_filename = ""
        input_file = None

        try:
            # Create the input file
            input_fd, input_filename = tempfile.mkstemp(suffix=".input", text=True)
            input_file = os.fdopen(input_fd, "w")
            input_file.write(file_contents)
            input_file.close()

            # Get the patch file name so it can be removed if it's created.
            # pymend will create it in the current working directory
            patch_filename = os.path.join(
                self.CWD, f"{os.path.basename(input_filename)}.patch"
            )

            cmd_args = f"{cmd_args} {input_filename}"

            if overwrite_mode:
                cmd_args = f"{cmd_args} --write "

            self.run_pymend_app_and_assert_is_expected(
                cmd_args=cmd_args,
                expected_stderr=expected_stderr,
                expected_returncode=expected_returncode,
                expected_stdout=expected_stdout,
                write_to_stdin=file_contents,
            )

            if overwrite_mode:
                with open(input_filename, encoding="utf-8") as file:
                    output = file.read()
            else:
                with open(patch_filename, encoding="utf-8") as file:
                    output = file.read()
                # The expected output will have filenames of '-'
                # - replace them with the actual filename
                output = re.sub(
                    rf"/{os.path.basename(input_filename)}$",
                    r"/-",
                    output,
                    flags=re.MULTILINE,
                )

            normalised_output = self.normalise_empty_lines(output)
            normalised_expected_output = self.normalise_empty_lines(
                expected_file_contents
            )
            assert remove_diff_header(normalised_output) == remove_diff_header(
                normalised_expected_output
            ), (
                f"Output from cmd: {cmd_args} was:\n{normalised_output!r}\nnot "
                f"the expected:\n{normalised_expected_output!r}"
            )

        finally:
            if input_filename:
                if input_file and not input_file.closed:
                    input_file.close()
                os.remove(input_filename)

            if not overwrite_mode and os.path.isfile(patch_filename):
                os.remove(patch_filename)

    def test_overwrite_files_the_same(self) -> None:
        """Test that the file is correct when the output is the same as the input."""
        self.run_pymend_app_with_a_file_and_assert_is_expected(
            file_contents=self.EXPECTED_OUTPUT,
            expected_stderr=re.compile("All done! .*"),
            expected_file_contents=self.EXPECTED_OUTPUT,
            overwrite_mode=True,
        )

    def test_overwrite_files_different(self) -> None:
        """Test the file is overwritten with the correct result."""
        self.run_pymend_app_with_a_file_and_assert_is_expected(
            file_contents=self.INPUT,
            expected_file_contents=self.EXPECTED_OUTPUT,
            expected_stderr=re.compile("All done! .*"),
            expected_stdout=re.compile(
                r"Modified docstrings of elements \(Module, func\) in file.*", re.DOTALL
            ),
            overwrite_mode=True,
        )

    def test_patch_files_the_same(self) -> None:
        """Check the patch file created when the files are the same."""
        with pytest.raises(FileNotFoundError):
            self.run_pymend_app_with_a_file_and_assert_is_expected(
                file_contents=self.EXPECTED_OUTPUT,
                expected_stderr=re.compile("All done! .*"),
                expected_file_contents=self.PATCH_PREFIX + "\n",
            )

    def test_patch_files_different(self) -> None:
        """Test the patch file is correct."""
        self.run_pymend_app_with_a_file_and_assert_is_expected(
            file_contents=self.INPUT,
            expected_stderr=re.compile("All done! .*"),
            expected_file_contents=self.EXPECTED_PATCH,
        )



"""Test general pymend functionality."""


import os
import pathlib
import re
import shutil

import pytest

import pymend.pymend as pym
from pymend.types import FixerSettings

CURRENT_DIR = os.path.dirname(__file__)


def absdir(file: str) -> str:
    """Get absolute path for file.

    Parameters
    ----------
    file : str
        File path

    Returns
    -------
    str
        Absolute path to file
    """
    return os.path.join(CURRENT_DIR, file)


def get_expected_patch(name: str) -> str:
    """Open a patch file, and if found Pymend signature remove the 2 first lines.

    Parameters
    ----------
    name : str
        Name of the patch

    Returns
    -------
    str
        Expected patch as a string.
    """
    try:
        with open(absdir(f"refs/{name}"), encoding="utf-8") as file:
            expected_lines = file.readlines()
            if expected_lines[0].startswith("# Patch"):
                expected_lines = expected_lines[2:]
            expected = "".join(expected_lines)
    except Exception as error:  # noqa: BLE001
        pytest.fail(f'Raised exception: "{error}"')
    return expected


class TestDocStrings:
    """Test correct parsing of docstrings."""

    def setup_class(self) -> None:
        """Set up class by setting file paths."""
        self.myelem = '    def my_method(self, first, second=None, third="value"):'
        self.mydocs = '''        """This is a description of a method.
                It is on several lines.
                Several styles exists:
                    -javadoc,
                    -reST,
                    -cstyle.
                It uses the javadoc style.

                @param first: the 1st argument.
                with multiple lines
                @type first: str
                @param second: the 2nd argument.
                @return: the result value
                @rtype: int
                @raise KeyError: raises exception

                """'''

        self.inifile = absdir("refs/origin_test.py")
        self.jvdfile = absdir("refs/javadoc_test.py")
        self.rstfile = absdir("refs/rest_test.py")
        self.foo = absdir("refs/foo")

        # prepare test file
        txt = ""
        shutil.copyfile(self.inifile, self.jvdfile)
        txt = pathlib.Path(self.jvdfile).read_text()
        txt = txt.replace("@return", ":returns")
        txt = txt.replace("@raise", ":raises")
        txt = txt.replace("@", ":")
        with open(self.rstfile, "w", encoding="utf-8") as rstfile:
            rstfile.write(txt)
        with open(self.foo, "w", encoding="utf-8") as fooo:
            fooo.write("foo")

    def teardown_class(self) -> None:
        """Tear down class by deleting files."""
        os.remove(self.jvdfile)
        os.remove(self.rstfile)
        os.remove(self.foo)

    def test_parsed_javadoc(self) -> None:
        """Test that javadoc comments get parsed."""
        comment = pym.PyComment(self.inifile, fixer_settings=FixerSettings())
        assert comment.fixed

    def test_windows_rename(self) -> None:
        """Check that renaming works correctly."""
        bar = absdir("bar")
        with open(bar, "w", encoding="utf-8") as fbar:
            fbar.write("bar")
        comment = pym.PyComment(self.foo, fixer_settings=FixerSettings())
        comment._windows_rename(bar)
        assert not os.path.isfile(bar)
        assert os.path.isfile(self.foo)
        foo_txt = pathlib.Path(self.foo).read_text(encoding="utf-8")
        assert foo_txt == "bar"


def remove_diff_header(diff: str) -> str:
    """Remove header differences from diff.

    Parameters
    ----------
    diff : str
        Diff file to clean.

    Returns
    -------
    str
        Cleaned diff.
    """
    return re.sub(r"(@@.+@@)|(\-\-\-.*)|(\+\+\+.*)", "", diff)


class TestFilesConversions:
    """Test patch files."""

    def test_case_free_testing(self) -> None:
        """Test correct handling for case where input style in ambiguous."""
        comment = pym.PyComment(
            absdir("refs/free_cases.py"), fixer_settings=FixerSettings()
        )
        result = "".join(comment._docstring_diff())
        assert result == ""

    def test_case_gen_all_params_numpydoc(self) -> None:
        """Test generation of numpydoc patch.

        The file has several functions with no or several parameters,
        so Pymend should produce docstrings in numpydoc format.
        """
        expected = get_expected_patch("params.py.patch.numpydoc.expected")
        comment = pym.PyComment(
            absdir("refs/params.py"), fixer_settings=FixerSettings()
        )
        result = "".join(comment._docstring_diff())
        assert remove_diff_header(result) == remove_diff_header(expected)

    def test_case_no_gen_docs_already_numpydoc(self) -> None:
        """Test that correct format needs no fixing.

        The file has functions with already docstrings in numpydoc format,
        so no docstring should be produced.
        """
        comment = pym.PyComment(
            absdir("refs/docs_already_numpydoc.py"), fixer_settings=FixerSettings()
        )
        result = "".join(comment._docstring_diff())
        assert result == ""



"""Utility functions for working with docstrings."""
from collections import ChainMap
from collections.abc import Iterable
from inspect import Signature
from itertools import chain
from typing import Callable

from .common import (
    DocstringMeta,
    DocstringParam,
    DocstringStyle,
    RenderingStyle,
)
from .parser import compose, parse

_Func = Callable[..., object]


def combine_docstrings(
    *others: _Func,
    exclude: Iterable[type[DocstringMeta]] = (),
    style: DocstringStyle = DocstringStyle.AUTO,
    rendering_style: RenderingStyle = RenderingStyle.COMPACT,
) -> _Func:
    """Combine docstrings of multiple functions.

    Parses the docstrings from `others`,
    programmatically combines them with the parsed docstring of the decorated
    function, and replaces the docstring of the decorated function with the
    composed result. Only parameters that are part of the decorated functions
    signature are included in the combined docstring. When multiple sources for
    a parameter or docstring metadata exists then the decorator will first
    default to the wrapped function's value (when available) and otherwise use
    the rightmost definition from ``others``.

    Parameters
    ----------
    *others : _Func
        callables from which to parse docstrings.
    exclude : Iterable[type[DocstringMeta]]
        an iterable of ``DocstringMeta`` subclasses to exclude when
        combining docstrings. (Default value = ())
    style : DocstringStyle
        Style that the docstrings are currently in. Default will infer style.
        (Default value = DocstringStyle.AUTO)
    rendering_style : RenderingStyle
        Rendering style to use. (Default value = RenderingStyle.COMPACT)

    Returns
    -------
    _Func
        the decorated function with a modified docstring.

    Examples
    --------
    >>> def fun1(a, b, c, d):
    ...    '''short_description: fun1
    ...
    ...    :param a: fun1
    ...    :param b: fun1
    ...    :return: fun1
    ...    '''
    >>> def fun2(b, c, d, e):
    ...    '''short_description: fun2
    ...
    ...    long_description: fun2
    ...
    ...    :param b: fun2
    ...    :param c: fun2
    ...    :param e: fun2
    ...    '''
    >>> @combine_docstrings(fun1, fun2)
    >>> def decorated(a, b, c, d, e, f):
    ...     '''
    ...     :param e: decorated
    ...     :param f: decorated
    ...     '''
    >>> print(decorated.__doc__)
    short_description: fun2
    <BLANKLINE>
    long_description: fun2
    <BLANKLINE>
    :param a: fun1
    :param b: fun1
    :param c: fun2
    :param e: fun2
    :param f: decorated
    :returns: fun1
    >>> @combine_docstrings(fun1, fun2, exclude=[DocstringReturns])
    >>> def decorated(a, b, c, d, e, f): pass
    >>> print(decorated.__doc__)
    short_description: fun2
    <BLANKLINE>
    long_description: fun2
    <BLANKLINE>
    :param a: fun1
    :param b: fun1
    :param c: fun2
    :param e: fun2
    """

    def wrapper(func: _Func) -> _Func:
        """Wrap the function.

        Parameters
        ----------
        func : _Func
            Function to wrap.

        Returns
        -------
        _Func
            Wrapped function
        """
        sig = Signature.from_callable(func)

        comb_doc = parse(func.__doc__ or "")
        docs = [parse(other.__doc__ or "") for other in others] + [comb_doc]
        params = dict(
            ChainMap(*({param.arg_name: param for param in doc.params} for doc in docs))
        )

        for doc in reversed(docs):
            if not doc.short_description:
                continue
            comb_doc.short_description = doc.short_description
            comb_doc.blank_after_short_description = doc.blank_after_short_description
            break

        for doc in reversed(docs):
            if not doc.long_description:
                continue
            comb_doc.long_description = doc.long_description
            comb_doc.blank_after_long_description = doc.blank_after_long_description
            break

        combined: dict[type[DocstringMeta], list[DocstringMeta]] = {}
        for doc in docs:
            metas: dict[type[DocstringMeta], list[DocstringMeta]] = {}
            for meta in doc.meta:
                meta_type = type(meta)
                if meta_type in exclude:
                    continue
                metas.setdefault(meta_type, []).append(meta)
            for meta_type, meta in metas.items():
                combined[meta_type] = meta

        combined[DocstringParam] = [
            params[name] for name in sig.parameters if name in params
        ]
        comb_doc.meta = list(chain(*combined.values()))
        func.__doc__ = compose(comb_doc, style=style, rendering_style=rendering_style)
        return func

    return wrapper



"""Common methods for parsing."""
import enum
from dataclasses import dataclass
from typing import Optional, Union

from typing_extensions import TypeAlias

PARAM_KEYWORDS = {
    "param",
    "parameter",
    "arg",
    "argument",
    "attribute",
    "key",
    "keyword",
}
# These could be made into frozen sets
# Then one could create a dictionary
# {KEYWORD: FUNCTION_THAT_HANDLES_THAT_KEYWORD_SECTION}  # noqa: ERA001
RAISES_KEYWORDS = {"raises", "raise", "except", "exception"}
DEPRECATION_KEYWORDS = {"deprecation", "deprecated"}
RETURNS_KEYWORDS = {"return", "returns"}
YIELDS_KEYWORDS = {"yield", "yields"}
EXAMPLES_KEYWORDS = {"example", "examples"}


def clean_str(string: str) -> Optional[str]:
    """Strip a string and return None if it is now empty.

    Parameters
    ----------
    string : str
        String to clean

    Returns
    -------
    Optional[str]
        None of the stripped string is empty. Otherwise the stripped string.
    """
    string = string.strip()
    return string if string != "" else None


class ParseError(RuntimeError):
    """Base class for all parsing related errors."""


class DocstringStyle(enum.Enum):
    """Docstring style."""

    REST = 1
    GOOGLE = 2
    NUMPYDOC = 3
    EPYDOC = 4
    AUTO = 255


class RenderingStyle(enum.Enum):
    """Rendering style when unparsing parsed docstrings."""

    COMPACT = 1
    CLEAN = 2
    EXPANDED = 3


@dataclass
class DocstringMeta:
    """Docstring meta information.

    Symbolizes lines in form of

    Parameters
    ----------
    args : List[str]
        list of arguments. The exact content of this variable is
        dependent on the kind of docstring; it's used to distinguish
        between custom docstring meta information items.
    description : Optional[str]
        associated docstring description.
    """

    args: list[str]
    description: Optional[str]


@dataclass
class DocstringParam(DocstringMeta):
    """DocstringMeta symbolizing :param metadata."""

    arg_name: str
    type_name: Optional[str]
    is_optional: Optional[bool]
    default: Optional[str]


@dataclass
class DocstringReturns(DocstringMeta):
    """DocstringMeta symbolizing :returns metadata."""

    type_name: Optional[str]
    is_generator: bool
    return_name: Optional[str] = None


@dataclass
class DocstringYields(DocstringMeta):
    """DocstringMeta symbolizing :yields metadata."""

    type_name: Optional[str]
    is_generator: bool
    yield_name: Optional[str] = None


@dataclass
class DocstringRaises(DocstringMeta):
    """DocstringMeta symbolizing :raises metadata."""

    type_name: Optional[str]
    description: Optional[str]


MainSections: TypeAlias = Union[
    DocstringParam, DocstringRaises, DocstringReturns, DocstringYields
]


@dataclass
class DocstringDeprecated(DocstringMeta):
    """DocstringMeta symbolizing deprecation metadata."""

    version: Optional[str]


@dataclass
class DocstringExample(DocstringMeta):
    """DocstringMeta symbolizing example metadata."""

    snippet: Optional[str]


class Docstring:
    """Docstring object representation."""

    def __init__(
        self,
        style: Optional[DocstringStyle] = None,
    ) -> None:
        """Initialize self.

        Parameters
        ----------
        style : Optional[DocstringStyle]
            Style that this docstring was formatted in. (Default value = None)
        """
        self.short_description: Optional[str] = None
        self.long_description: Optional[str] = None
        self.blank_after_short_description: bool = False
        self.blank_after_long_description: bool = False
        self.meta: list[DocstringMeta] = []
        self.style: Optional[DocstringStyle] = style

    @property
    def params(self) -> list[DocstringParam]:
        """Return a list of information on function params.

        Returns
        -------
        list[DocstringParam]
            list of information on function params
        """
        return [item for item in self.meta if isinstance(item, DocstringParam)]

    @property
    def raises(self) -> list[DocstringRaises]:
        """Return a list of the exceptions that the function may raise.

        Returns
        -------
        list[DocstringRaises]
            list of the exceptions that the function may raise.
        """
        return [item for item in self.meta if isinstance(item, DocstringRaises)]

    @property
    def returns(self) -> Optional[DocstringReturns]:
        """Return a single information on function return.

        Takes the first return information.

        Returns
        -------
        Optional[DocstringReturns]
            Single information on function return.
        """
        return next(
            (item for item in self.meta if isinstance(item, DocstringReturns)),
            None,
        )

    @property
    def many_returns(self) -> list[DocstringReturns]:
        """Return a list of information on function return.

        Returns
        -------
        list[DocstringReturns]
            list of information on function return.
        """
        return [item for item in self.meta if isinstance(item, DocstringReturns)]

    @property
    def yields(self) -> Optional[DocstringYields]:
        """Return information on function yield.

        Takes the first generator information.

        Returns
        -------
        Optional[DocstringYields]
            Single information on function yield.
        """
        return next(
            (
                item
                for item in self.meta
                if isinstance(item, DocstringYields) and item.is_generator
            ),
            None,
        )

    @property
    def many_yields(self) -> list[DocstringYields]:
        """Return a list of information on function yields.

        Returns
        -------
        list[DocstringYields]
            list of information on function yields.
        """
        return [item for item in self.meta if isinstance(item, DocstringYields)]

    @property
    def deprecation(self) -> Optional[DocstringDeprecated]:
        """Return a single information on function deprecation notes.

        Returns
        -------
        Optional[DocstringDeprecated]
            single information on function deprecation notes.
        """
        return next(
            (item for item in self.meta if isinstance(item, DocstringDeprecated)),
            None,
        )

    @property
    def examples(self) -> list[DocstringExample]:
        """Return a list of information on function examples.

        Returns
        -------
        list[DocstringExample]
            list of information on function examples.
        """
        return [item for item in self.meta if isinstance(item, DocstringExample)]


def split_description(docstring: Docstring, desc_chunk: str) -> None:
    """Break description into short and long parts.

    Parameters
    ----------
    docstring : Docstring
        Docstring to fill with description information.
    desc_chunk : str
        Chunk of the raw docstring representing the description.
    """
    parts = desc_chunk.split("\n", 1)
    docstring.short_description = parts[0] or None
    if len(parts) > 1:
        long_desc_chunk = parts[1] or ""
        docstring.blank_after_short_description = long_desc_chunk.startswith("\n")
        docstring.blank_after_long_description = long_desc_chunk.endswith("\n\n")
        docstring.long_description = long_desc_chunk.strip() or None


def append_description(docstring: Docstring, parts: list[str]) -> None:
    """Append the docstrings description to the output stream.

    Parameters
    ----------
    docstring : Docstring
        Docstring whose information should be added.
    parts : list[str]
        List of strings representing the output of compose().
        Descriptions should be added to this.
    """
    if docstring.short_description:
        parts.append(docstring.short_description)
    if docstring.blank_after_short_description:
        parts.append("")
    if docstring.long_description:
        parts.append(docstring.long_description)
    if docstring.blank_after_long_description:
        parts.append("")



"""_summary_."""
def func1(param1):
    """Function 1 with 1 param.

    Parameters
    ----------
    param1 : type
        1st parameter

    Returns
    -------
    string
        a message

    Raises
    ------
    KeyError
        when a key error
    """
    return "huh"


def func2(param1, param2):
    """Function 2 with 2 params.

    Parameters
    ----------
    param1 : type
        1st parameter
    param2 : type
        2nd parameter
    """
    pass





"""Unit tests for pymend.types."""


import pytest

from pymend.types import Parameter


class TestParameter:
    """Test the Parameter dataclass."""

    @pytest.mark.parametrize(
        ("source", "expected"),
        [
            (
                [
                    Parameter("3"),
                    Parameter("1"),
                    Parameter("1"),
                    Parameter("4"),
                    Parameter("5"),
                    Parameter("6"),
                    Parameter("6"),
                    Parameter("3"),
                ],
                [
                    Parameter("3"),
                    Parameter("1"),
                    Parameter("4"),
                    Parameter("5"),
                    Parameter("6"),
                ],
            ),
        ],
    )
    def test_uniquefy(self, source: list[Parameter], expected: list[Parameter]) -> None:
        """Test that uniquefy removes duplicates and keeps order."""
        assert list(Parameter.uniquefy(source)) == expected



"""_summary_."""


def ex1(
    docstring: str,
    quotes: str = '"""',
) -> None:
    pass


def ex2(
    docstring: str,
    quotes: str = '"""',
) -> None:
    """_summary_.

    Parameters
    ----------
    docstring : str
        The input docstring
    quotes : str
        The quotes to use.
    """
    print(x)

def ex3(
    docstring: str,
    quotes: str = '"""',
) -> None:
    """_summary_.

    Parameters
    ----------
    docstring : str
        The input docstring
    quotes : str
        The quotes to use. (Default value = '\"\"\"')
    """
    print(x)



"""_summary_."""

def test1(param: int):
    """Short summary.
    Parameters
    ----------
    param : int
        description.
    """
    pass

def test2(param: int):
    """Short summary.

    Long description
    Parameters
    ----------
    param : int
        description.
    """
    pass

def test3(param: int):
    """Short summary.

    """
    pass

def test4(param: int):
    """Short summary.

    Long description

    


    """
    pass



"""_summary_."""

def raises():
    """_summary_.

    Raises
    ------
    ValueError
        _description_
    TypeError
        _description_
    """
    raise ValueError
    raise KeyError
    raise TypeError(msg)
    raise ValueError